// ===================================================================
// $Id: $
//
// CleanMesh.cpp
//
// Class: CCleanMesh
//
//  The source file to implement rendering via OpenGL with shaders
//
// Initial coding by Vlastimil Havran, 2013.

#include "pgr.h"
#include <GL/freeglut.h>
//#include <GL/glext.h>
#include <GL/gl.h>
#include <GL/glu.h>

// in order to get function prototypes from glext.h, define
//#GL_GLEXT_PROTOTYPES before including glext.h
//#define GL_GLEXT_PROTOTYPES

#include "glext.h"
//#include "glInfo.h"

#include "pgr.h"
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <vector>
#include <cassert>
#include <cstring>
#include <sys/stat.h>
#include <cmath>
#include <iostream>
#include <iostream>
#include <sstream>
#include <string>
#include <iomanip>

using namespace std;

#include "configg.h"
#include "bconfig.h"
#include "basstr.h"
#include "hdrloader.h"
#include "CleanMesh.h"

static std::string vertexShader3 =
"#version 140\n"
"uniform mat4 PVMmatrix;\n"
"in vec3 vertex; // input vertex\n"
"// in vec3 normal; // input normal\n"
"out vec3 position_v;\n"
"// out vec3 normal_v;\n"
"void main() {\n"
"  gl_Position = PVMmatrix * vec4(vertex, 1);\n"
"}\n"
;

static std::string fragmentShader3 =
"#version 140\n"
"//uniform mat4 Vmatrix;         // view (camera) transform --> world to eye coordinates\n"
"uniform vec3 ccolor; // color\n"
"out vec4 outputColor;\n"
"void main() {\n"
"  outputColor = vec4(ccolor.x, ccolor.y, ccolor.z, 0);\n"
"}\n"
;

// initialize shaders
bool
CCleanMesh::createShaderProgramNoNormals()
{
  // load shader
  std::vector<GLuint> shaders;
  // load vertex shader
  int glIDvs = pgr::createShaderFromSource(GL_VERTEX_SHADER, vertexShader3);
  if (glIDvs > 0)
    shaders.push_back(glIDvs);
  else {
    cout << "ERROR" << endl;
    return true; // error
  }
  // load fragment shader
  int glIDfs = pgr::createShaderFromSource(GL_FRAGMENT_SHADER, fragmentShader3);
  if (glIDfs > 0)
    shaders.push_back(glIDfs);
  else {
    cout << "ERROR" << endl;
    return true; // error
  }

  // generate program to run
  shaderProgram = pgr::createProgram(shaders);
  //cout << "Init MeshShaderProgram = " << shaderProgram << endl;
  //assert(shaderProgram > -1);
  return false; // OK
}

void
CCleanMesh::setRenderMode(int renderModeV)
{
  if (!((renderModeV == GL_POINTS)||
	(renderModeV == GL_LINE_STRIP)||
	(renderModeV == GL_LINE_LOOP)||
	(renderModeV == GL_LINE_LOOP) ||
	(renderModeV == GL_LINES) ||
	(renderModeV == GL_LINE_STRIP_ADJACENCY) ||
	(renderModeV == GL_LINES_ADJACENCY) ||
	(renderModeV == GL_TRIANGLE_STRIP) ||
	(renderModeV == GL_TRIANGLE_FAN) ||
	(renderModeV == GL_TRIANGLES) ||
	(renderModeV == GL_TRIANGLE_STRIP_ADJACENCY) ||
	(renderModeV == GL_TRIANGLES_ADJACENCY))) {
    cout << "Wrong rendering mode = " << renderModeV << endl;
    exit(-1);
  }
  // copy render mode
  renderMode = renderModeV;
  return;
}

// 1 .. convert to lines, 0 .. convert to triangles from lines
void
CCleanMesh::toggleLinesTriangles(int mode)
{
#if 0
  cout << "BEFORE" << endl;
  cout << "cntVertices = " << cntVertices << endl;
  for (int i = 0; i < cntVertices*3; i+=3) {
    cout << vertices[i] << " " << vertices[i+1] << " " << vertices[i+2] << endl;
  }
#endif  
  if (mode) {
    float *vv = new float[cntVertices*3*2]; // each vertex needs 3 floats
    assert(vv);
    for (int i = 0; i < cntVertices*3; i += 9) { // one triangle at a time
      int d = 2*i; // destination
      int s = i; // src
      // first line segment
      copyVertex(vv, vertices, d, s); 
      copyVertex(vv, vertices, d+3, s+3); 
      // second line segment
      copyVertex(vv, vertices, d+6, s+3); 
      copyVertex(vv, vertices, d+9, s+6); 
      // third line segment
      copyVertex(vv, vertices, d+12, s+6); 
      copyVertex(vv, vertices, d+15, s+0);
    }
    cntVertices *= 2; // the number of vertices is twice higher now
    delete []vertices;
    vertices = vv;
    glBindVertexArray(vaoHandle);
    CHECK_GL_ERROR();
    glBindBuffer(GL_ARRAY_BUFFER, vbovertices);
    CHECK_GL_ERROR();
    int sizeVertices = 3 * sizeof(float) * cntVertices; // each vertex 3 coordinates
    glBufferData(GL_ARRAY_BUFFER, sizeVertices, vertices, GL_STATIC_DRAW);
    CHECK_GL_ERROR();
    //cout << "sizeof(vertices) = " << sizeof(vertices) << endl;
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    CHECK_GL_ERROR();
    glBindVertexArray(0);
    CHECK_GL_ERROR();
  }
  else {
    cntVertices = cntVertices/2;
    float *vv = new float[cntVertices*3]; // each vertex needs 3 floats
    assert(vv);
    for (int i = 0; i < cntVertices*3; i += 9) { // three line segments at a time
      int d = i; // destination
      int s = 2*i; // src
      // triangles from 3 segments
      copyVertex(vv, vertices, d, s); 
      copyVertex(vv, vertices, d+3, s+6); 
      copyVertex(vv, vertices, d+6, s+12);
    }
    delete []vertices;
    vertices = vv;
    glBindVertexArray(vaoHandle);
    CHECK_GL_ERROR();
    glBindBuffer(GL_ARRAY_BUFFER, vbovertices);
    CHECK_GL_ERROR();
    int sizeVertices = 3 * sizeof(float) * cntVertices; // each vertex 3 coordinates
    glBufferData(GL_ARRAY_BUFFER, sizeVertices, vertices, GL_STATIC_DRAW);
    CHECK_GL_ERROR();
    //cout << "sizeof(vertices) = " << sizeof(vertices) << endl;
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    CHECK_GL_ERROR();
    glBindVertexArray(0);
    CHECK_GL_ERROR();
  }
#if 0
  cout << "cntVertices = " << cntVertices << endl;
  for (int i = 0; i < cntVertices*3; i+=3) {
    cout << vertices[i] << " " << vertices[i+1] << " " << vertices[i+2] << endl;
  }
#endif  
  return;
} 

void
CCleanMesh::init(int cntVerticesV, int renderModeV, float *verticesV)
{
  setRenderMode(renderModeV);  

  if (createShaderProgramNoNormals()) {
    cout << "No shader program can be created" << endl;
    exit(-1);
  }
  
  assert(verticesV);
  this->cntVertices = cntVerticesV;
  if (vertices) {
    delete [] vertices;
    vertices = 0;
  }
  vertices = new float[cntVertices*3]; // each vertex needs 3 floats
  assert(vertices);
  for (int i = 0; i < cntVertices*3; i++) vertices[i] = verticesV[i];
    
  // Input triangle mesh
  // handles to vertex shader inputs
  int maxattribs;
  glGetIntegerv(GL_MAX_VERTEX_ATTRIBS, &maxattribs);
  //cout << "maxVertexAttribs = " << maxattribs << endl;
  
  // buffer object for the triangles - vertices
  glGenVertexArrays(1, &vaoHandle);
  CHECK_GL_ERROR();
  glBindVertexArray(vaoHandle);
  CHECK_GL_ERROR();
  
  // array of vertex coordinates
  glGenBuffers(1, &vbovertices);
  glBindBuffer(GL_ARRAY_BUFFER, vbovertices);
  CHECK_GL_ERROR();
  int sizeVertices = 3 * sizeof(float) * cntVerticesV; // each vertex 3 coordinates
  glBufferData(GL_ARRAY_BUFFER, sizeVertices, vertices, GL_STATIC_DRAW);
  CHECK_GL_ERROR();
  //cout << "sizeof(vertices) = " << sizeof(vertices) << endl;
  glBindBuffer(GL_ARRAY_BUFFER, 0);
  //glDisableVertexAttribArray(vertexLoc);
  CHECK_GL_ERROR();
  vertexLoc = glGetAttribLocation(shaderProgram, "vertex");

  ccolor[0] = 1; // the color to be used
  ccolor[1] = 1; // the color to be used
  ccolor[2] = 1; // the color to be used
  
  return; // OK
}

void
CCleanMesh::free()
{
  //glDeleteBuffers(1, &vbonormals);
  if (vbovertices>0) glDeleteBuffers(1, &vbovertices);
  if (vaoHandle>0) glDeleteVertexArrays(1, &vaoHandle);
  delete []vertices;
  //delete []normals;
  pgr::deleteProgramAndShaders(shaderProgram);
}

void
CCleanMesh::draw(const glm::mat4 &PVMmatrix)
{
  // Binding right program and matrices
  //glUseProgram(0); //detach old program
  //CHECK_GL_ERROR();
  glUseProgram(shaderProgram);
  CHECK_GL_ERROR();
  glBindVertexArray(vaoHandle);
  CHECK_GL_ERROR();

  // Attribute variables
  //cout << "vertexLoc = " << vertexLoc << endl;
  //cout << "texLoc = " << texLoc << endl;

  glEnableVertexAttribArray(vertexLoc);
  CHECK_GL_ERROR();
  glBindBuffer(GL_ARRAY_BUFFER, vbovertices);
  CHECK_GL_ERROR();
  glVertexAttribPointer(vertexLoc, 3, GL_FLOAT, GL_FALSE, 0, (GLubyte *)NULL);
  CHECK_GL_ERROR();
#if 0
  // Normal
  glEnableVertexAttribArray(normalLoc);  
  glBindBuffer(GL_ARRAY_BUFFER, vbonormals);
  glVertexAttribPointer(normalLoc, 3, GL_FLOAT, GL_FALSE, 0, (GLubyte *)NULL);
#endif

  // Uniform variables for the matrix
  GLint modelMatLoc = glGetUniformLocation(shaderProgram, "PVMmatrix");
  CHECK_GL_ERROR();
  // binding matrix to shader
  glUniformMatrix4fv(modelMatLoc, 1, GL_FALSE, glm::value_ptr(PVMmatrix)); 
  CHECK_GL_ERROR();
  GLint fragColorLoc = glGetUniformLocation(shaderProgram, "ccolor");
  CHECK_GL_ERROR();
  if (fragColorLoc < 1) {
    cout << "ccolor not found" << endl;
    exit(0);
  }
  glUniform3fv(fragColorLoc, 1, ccolor); // binding matrix to shader
  CHECK_GL_ERROR();
  // Binding vertex array object
  glBindVertexArray(vaoHandle);
  glDrawArrays(renderMode, 0, cntVertices);
  CHECK_GL_ERROR();
  glBindVertexArray(0); // detach
  CHECK_GL_ERROR();
  glUseProgram(0);
  CHECK_GL_ERROR();
  return;
}

void
CCleanMesh::rotate3DPoint(float angle, float *v, int axis)
{
  float sv[3];
  // ... rotation of whole enviroment sphere by angle phiShift
  int a,b;
  if(axis == 0) {
    a = 0;
    b = 1;
  }
  if(axis == 1) {
    a = 0;
    b = 2;
  }
  if(axis == 2) {
    a = 1;
    b = 2;
  }

  sv[a] = cos(angle) * v[a] - sin(angle) * v[b];
  sv[b] = sin(angle) * v[a] + cos(angle) * v[b];
  v[a] = sv[a];
  v[b] = sv[b];
  return;
}//-- rotate3DPoint -------------------------------------------------------

// rotate object geometry by an angle around axis
void
CCleanMesh::rotateOBJModelGeometry(int axis, float rotAngle)
{
  for(int f=0;f < cntVertices; f+=3) { // for all triangles
    rotate3DPoint(rotAngle, &(vertices[f]), axis);    
  } // for f
  
  glDeleteBuffers(1, &vbovertices);
  glGenBuffers(1, &vbovertices);
  glBindBuffer(GL_ARRAY_BUFFER, vbovertices);
  //glBindVertexArray(vbovertices);
  CHECK_GL_ERROR();
  int sizeVertices = 3 * sizeof(float) * cntVertices; // each vertex 3 coordinates
  glBufferData(GL_ARRAY_BUFFER, sizeVertices, vertices, GL_STATIC_DRAW);
  CHECK_GL_ERROR();
  //cout << "sizeof(vertices) = " << sizeof(vertices) << endl;
  glBindBuffer(GL_ARRAY_BUFFER, 0);
  //glBindVertexArray(0);
  return;
}//--- rotateOBJModelGeometry -----------------------------------------------

// translate object geometry by a shift in some axis
void
CCleanMesh::translateOBJModelGeometry(int axis, float value)
{
  for(int f=0;f < cntVertices; f+=3)
  {// for faces
    vertices[f+axis] += value;
  }
  //glBindBuffer(GL_ARRAY_BUFFER, vbovertices);
  glDeleteBuffers(1, &vbovertices);
  glGenBuffers(1, &vbovertices);
  glBindVertexArray(vbovertices);
  CHECK_GL_ERROR();
  int sizeVertices = 3 * sizeof(float) * cntVertices; // each vertex 3 coordinates
  glBufferData(GL_ARRAY_BUFFER, sizeVertices, vertices, GL_STATIC_DRAW);
  CHECK_GL_ERROR();
  //cout << "sizeof(vertices) = " << sizeof(vertices) << endl;
  //glBindBuffer(GL_ARRAY_BUFFER, 0);
  glBindVertexArray(0);
  return;  
}

