// ===================================================================
// $Id: $
//
// TexturedMesh.cpp
//
// Class: CTexturedMesh
//
//  The header file to implement rendering via OpenGL with shaders and HDR texture
//
// Initial coding by Vlastimil Havran, 2013.

#include "pgr.h"
#include <GL/freeglut.h>
//#include <GL/glext.h>
#include <GL/gl.h>
#include <GL/glu.h>

// in order to get function prototypes from glext.h, define
//#GL_GLEXT_PROTOTYPES before including glext.h
//#define GL_GLEXT_PROTOTYPES

#include "glext.h"
//#include "glInfo.h"

#include "pgr.h"
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <vector>
#include <cassert>
#include <cstring>
#include <sys/stat.h>
#include <cmath>
#include <iostream>
#include <iostream>
#include <sstream>
#include <string>
#include <iomanip>

using namespace std;

#include "configg.h"
#include "bconfig.h"
#include "basstr.h"
#include "hdrloader.h"
#include "TexturedMesh.h"

static std::string vertexShader =
"#version 140\n"
"uniform mat4 PVMmatrix;\n"
"in vec3 vertex; // input vertex\n"
"in vec2 texcoords; // input texture coordinate\n"
"// in vec3 normal; // input normal\n"
"out vec3 position_v;\n"
"smooth out vec2 vTexCoord_v; // texture coordinates\n"
"// out vec3 normal_v;\n"
"void main() {\n"
"  gl_Position = PVMmatrix * vec4(vertex, 1);\n"
"  vTexCoord_v = texcoords;\n"
"  //normal_v = modelMat * vec4(normal, 0.0f);\n"
"}\n"
;

static std::string fragmentShader =
"#version 140\n"
"//uniform mat4 Vmatrix;         // view (camera) transform --> world to eye coordinates\n"
"uniform sampler2D texsampler; // samples to access texture\n"
"smooth in vec2 vTexCoord_v;\n"
"out vec4 outputColor;\n"
"void main() {\n"
"  outputColor = texture(texsampler, vTexCoord_v);\n"
"}\n"
;

// initialize shaders
bool
CTexturedMesh::createShaderProgramNoNormals()
{
  // load shader
  std::vector<GLuint> shaders;
  // load vertex shader
  //int glIDvs = pgr::createShaderFromFile(GL_VERTEX_SHADER, "renderenv.vs");
  int glIDvs = pgr::createShaderFromSource(GL_VERTEX_SHADER, vertexShader);
  if (glIDvs > 0)
    shaders.push_back(glIDvs);
  else {
    cout << "ERROR" << endl;
    return true; // error
  }
  // load fragment shader
  //int glIDfs = pgr::createShaderFromFile(GL_FRAGMENT_SHADER, "renderenv.fs");
  int glIDfs = pgr::createShaderFromSource(GL_FRAGMENT_SHADER, fragmentShader);
  if (glIDfs > 0)
    shaders.push_back(glIDfs);
  else {
    cout << "ERROR" << endl;
    return true; // error
  }

  // generate program to run
  shaderProgram = pgr::createProgram(shaders);
  cout << "Init MeshShaderProgram = " << shaderProgram << endl;
  //assert(shaderProgram > -1);
  return false; // OK
}

void
CTexturedMesh::setRenderMode(int renderModeV)
{
  if (!((renderModeV == GL_POINTS)||
	(renderModeV == GL_LINE_STRIP)||
	(renderModeV == GL_LINE_LOOP)||
	(renderModeV == GL_LINE_LOOP) ||
	(renderModeV == GL_LINES) ||
	(renderModeV == GL_LINE_STRIP_ADJACENCY) ||
	(renderModeV == GL_LINES_ADJACENCY) ||
	(renderModeV == GL_TRIANGLE_STRIP) ||
	(renderModeV == GL_TRIANGLE_FAN) ||
	(renderModeV == GL_TRIANGLES) ||
	(renderModeV == GL_TRIANGLE_STRIP_ADJACENCY) ||
	(renderModeV == GL_TRIANGLES_ADJACENCY))) {
    cout << "Wrong rendering mode = " << renderModeV << endl;
    exit(-1);
  }
  // copy render mode
  renderMode = renderModeV;
  return;
}

// 1 .. convert to lines, 0 .. convert to triangles from lines
void
CTexturedMesh::toggleLinesTriangles(int mode)
{
#if 0
  cout << "BEFORE" << endl;
  cout << "cntVertices = " << cntVertices << endl;
  for (int i = 0; i < cntVertices*3; i+=3) {
    cout << vertices[i] << " " << vertices[i+1] << " " << vertices[i+2] << endl;
  }
#endif  
  if (mode) {
    float *vv = new float[cntVertices*3*2]; // each vertex needs 3 floats
    assert(vv);
    for (int i = 0; i < cntVertices*3; i += 9) { // one triangle at a time
      int d = 2*i; // destination
      int s = i; // src
      // first line segment
      copyVertex(vv, vertices, d, s); 
      copyVertex(vv, vertices, d+3, s+3); 
      // second line segment
      copyVertex(vv, vertices, d+6, s+3); 
      copyVertex(vv, vertices, d+9, s+6); 
      // third line segment
      copyVertex(vv, vertices, d+12, s+6); 
      copyVertex(vv, vertices, d+15, s+0);
    }
    cntVertices *= 2; // the number of vertices is twice higher now
    delete []vertices;
    vertices = vv;
    glBindVertexArray(vaoHandle);
    CHECK_GL_ERROR();
    glBindBuffer(GL_ARRAY_BUFFER, vbovertices);
    CHECK_GL_ERROR();
    int sizeVertices = 3 * sizeof(float) * cntVertices; // each vertex 3 coordinates
    glBufferData(GL_ARRAY_BUFFER, sizeVertices, vertices, GL_STATIC_DRAW);
    CHECK_GL_ERROR();
    //cout << "sizeof(vertices) = " << sizeof(vertices) << endl;
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    CHECK_GL_ERROR();
    glBindVertexArray(0);
    CHECK_GL_ERROR();
  }
  else {
    cntVertices = cntVertices/2;
    float *vv = new float[cntVertices*3]; // each vertex needs 3 floats
    assert(vv);
    for (int i = 0; i < cntVertices*3; i += 9) { // three line segments at a time
      int d = i; // destination
      int s = 2*i; // src
      // triangles from 3 segments
      copyVertex(vv, vertices, d, s); 
      copyVertex(vv, vertices, d+3, s+6); 
      copyVertex(vv, vertices, d+6, s+12);
    }
    delete []vertices;
    vertices = vv;
    glBindVertexArray(vaoHandle);
    CHECK_GL_ERROR();
    glBindBuffer(GL_ARRAY_BUFFER, vbovertices);
    CHECK_GL_ERROR();
    int sizeVertices = 3 * sizeof(float) * cntVertices; // each vertex 3 coordinates
    glBufferData(GL_ARRAY_BUFFER, sizeVertices, vertices, GL_STATIC_DRAW);
    CHECK_GL_ERROR();
    //cout << "sizeof(vertices) = " << sizeof(vertices) << endl;
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    CHECK_GL_ERROR();
    glBindVertexArray(0);
    CHECK_GL_ERROR();
  }
#if 0
  cout << "cntVertices = " << cntVertices << endl;
  for (int i = 0; i < cntVertices*3; i+=3) {
    cout << vertices[i] << " " << vertices[i+1] << " " << vertices[i+2] << endl;
  }
#endif  
  return;
} 

void
CTexturedMesh::init(int cntVerticesV, int renderModeV, float *verticesV, float *texcoordsV, int idTextureUnitV)
{
  setRenderMode(renderModeV);  

  if (createShaderProgramNoNormals()) {
    cout << "No shader program can be created" << endl;
    exit(-1);
  }
  this->idTextureUnit = idTextureUnitV;
  
  assert(verticesV);
  this->cntVertices = cntVerticesV;
  if (vertices) {
    delete [] vertices;
    vertices = 0;
  }
  vertices = new float[cntVertices*3]; // each vertex needs 3 floats
  assert(vertices);
  for (int i = 0; i < cntVertices*3; i++) vertices[i] = verticesV[i];
    
  // Input triangle mesh
  // handles to vertex shader inputs
  int maxattribs;
  glGetIntegerv(GL_MAX_VERTEX_ATTRIBS, &maxattribs);
  //cout << "maxVertexAttribs = " << maxattribs << endl;
  
  // buffer object for the triangles - vertices
  glGenVertexArrays(1, &vaoHandle);
  CHECK_GL_ERROR();
  glBindVertexArray(vaoHandle);
  CHECK_GL_ERROR();
  
  // array of vertex coordinates
  glGenBuffers(1, &vbovertices);
  glBindBuffer(GL_ARRAY_BUFFER, vbovertices);
  CHECK_GL_ERROR();
  int sizeVertices = 3 * sizeof(float) * cntVerticesV; // each vertex 3 coordinates
  glBufferData(GL_ARRAY_BUFFER, sizeVertices, vertices, GL_STATIC_DRAW);
  CHECK_GL_ERROR();
  //cout << "sizeof(vertices) = " << sizeof(vertices) << endl;
  glBindBuffer(GL_ARRAY_BUFFER, 0);
  //glDisableVertexAttribArray(vertexLoc);
  CHECK_GL_ERROR();
  vertexLoc = glGetAttribLocation(shaderProgram, "vertex");

  // Texture coordinates
  if (texcoords) {
    delete []texcoords;
    texcoords = 0;
  }

  if (texcoordsV) {
    texcoords = new float[cntVertices*2];
    assert(texcoords);
    for (int i = 0; i < cntVertices*2; i++) texcoords[i] = texcoordsV[i];
    // array of texture coordinates
    glGenBuffers(1, &vbotexcoords);
    glBindBuffer(GL_ARRAY_BUFFER, vbotexcoords);
    int sizeTexcoords = 2 * sizeof(float) * cntVerticesV;
    glBufferData(GL_ARRAY_BUFFER, sizeTexcoords, texcoords, GL_STATIC_DRAW);
    //cout << "sizeof(texcoords) = " << sizeof(texcoords) << endl;
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    CHECK_GL_ERROR();
    texLoc = glGetAttribLocation(shaderProgram, "texcoords");
  }
#if 0
  if (normals)
    delete []normals;
  if (normals) {
    // array of normals
    glGenBuffers(1, &vbonormals);
    glBindBuffer(GL_ARRAY_BUFFER, vbonormals);
    glBufferData(GL_ARRAY_BUFFER, sizeof(normals), normals, GL_STATIC_DRAW);  
    cout << "sizeof(normals) = " << sizeof(normals) << endl;
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    CHECK_GL_ERROR();
  }
#endif

  return; // OK
}

void
CTexturedMesh::loadTexture(const char *filename)
{
  // Floating point HDR texture created by reading the file
  CHDRImage hdrimage;
  cout << "Reading file HDR" << endl;
  if (!hdrimage.loadHDR(filename)) {
    cout << "The file " << filename << " cannot be read ... exiting" << endl;
    exit(-2);
  }
  if (texobject == 0) {
    glGenTextures(1, &texobject);
    CHECK_GL_ERROR();
  }
  glBindTexture(GL_TEXTURE_2D, texobject);
  CHECK_GL_ERROR();
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  CHECK_GL_ERROR();
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  CHECK_GL_ERROR();
  tex_width = hdrimage.getWidth();
  tex_height = hdrimage.getHeight();
  int cntsize = 4*tex_width*tex_height;
  if (texturedata) delete[]texturedata;
  texturedata = new float[cntsize];
  for (int i = 0; i < cntsize; i++) // copy data
    texturedata[i] = hdrimage.getBuf()[i];
  
  // copy the data to the texture
  glTexImage2D(GL_TEXTURE_2D, 0,  GL_RGBA32F,
	       tex_width, tex_height,
	       0, GL_RGBA, GL_FLOAT, &(texturedata[0]));
  CHECK_GL_ERROR();
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  CHECK_GL_ERROR();
  glBindTexture(GL_TEXTURE_2D, 0);
  CHECK_GL_ERROR();
  return;
}

void
CTexturedMesh::detachTextureData()
{
  //cout << "2texobject = " << texobject << endl;
  glDeleteTextures(1, &texobject);
  CHECK_GL_ERROR();
  texobject = 0;
  glBindTexture(GL_TEXTURE_2D, 0);
  CHECK_GL_ERROR();
  delete []texturedata;
  texturedata = 0;
}

void
CTexturedMesh::setTextureData(float *pointer, int tex_widthV, int tex_heightV)
{
  if (!pointer)
    return;
  
  detachTextureData();
  // buffer for texture created on a PC has some size
  int cntSize = 4 * tex_widthV * tex_heightV;
  if (texobject == 0) {
    glGenTextures(1, &texobject);
    //cout << "texobject = " << texobject << endl;
    CHECK_GL_ERROR();
  }
  glBindTexture(GL_TEXTURE_2D, texobject);
  CHECK_GL_ERROR();
  if ((tex_width != tex_widthV)||
      (tex_height != tex_heightV)||
      (texturedata == 0)) {
    tex_width = tex_widthV;
    tex_height = tex_heightV;
    if (texturedata) delete[]texturedata;
    texturedata = new float[cntSize];
    assert(texturedata);
  }
  assert(texturedata);
  for (int i = 0; i < cntSize; i++)
    texturedata[i] = pointer[i];
#if 0
  // setting some testing value
  for (int i = 0; i < cntSize; i++) {
    texturedata[i] = 1.f*(float)i/(float)cntSize;
  }
#endif  
#if 0
  // experimental setting some testing value
  for (int i = 0; i < cntSize; i +=4) {
    texturedata[i] = 0.0;
    texturedata[i+1] = 0.0;
    texturedata[i+2] = 0.0;
    texturedata[i+3] = 0.5;
  }
#endif  
  glTexImage2D(GL_TEXTURE_2D, 
	       0, 
	       GL_RGBA32F, // GL_RGBA32F, GL_RGB
	       tex_width, tex_height,
	       0, 
	       GL_RGBA, 
	       GL_FLOAT, 
	       &(texturedata[0]));
  CHECK_GL_ERROR();
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  CHECK_GL_ERROR();
  glBindTexture(GL_TEXTURE_2D, 0);
  CHECK_GL_ERROR();
  return; // OK
}

void
CTexturedMesh::free()
{
  //glDeleteBuffers(1, &vbonormals);
  if (vbotexcoords>0) glDeleteBuffers(1, &vbotexcoords);
  if (vbovertices>0) glDeleteBuffers(1, &vbovertices);
  if (vaoHandle>0) glDeleteVertexArrays(1, &vaoHandle);
  delete []vertices;
  delete []texcoords;
  delete []texturedata;
  //delete []normals;
  pgr::deleteProgramAndShaders(shaderProgram);
}

void
CTexturedMesh::draw(const glm::mat4 &PVMmatrix)
{
  // Binding right program and matrices
  //glUseProgram(0); //detach old program
  //CHECK_GL_ERROR();
  glUseProgram(shaderProgram);
  CHECK_GL_ERROR();
  glBindVertexArray(vaoHandle);
  CHECK_GL_ERROR();

  // Attribute variables
  //cout << "vertexLoc = " << vertexLoc << endl;
  //cout << "texLoc = " << texLoc << endl;

  glEnableVertexAttribArray(vertexLoc);
  CHECK_GL_ERROR();
  glBindBuffer(GL_ARRAY_BUFFER, vbovertices);
  CHECK_GL_ERROR();
  glVertexAttribPointer(vertexLoc, 3, GL_FLOAT, GL_FALSE, 0, (GLubyte *)NULL);
  CHECK_GL_ERROR();

  glEnableVertexAttribArray(texLoc);
  CHECK_GL_ERROR();
  glBindBuffer(GL_ARRAY_BUFFER, vbotexcoords);
  CHECK_GL_ERROR();
  glVertexAttribPointer(texLoc, 2, GL_FLOAT, GL_FALSE, 0, (GLubyte *)NULL);
  CHECK_GL_ERROR();

#if 0
  // Normal
  glEnableVertexAttribArray(normalLoc);  
  glBindBuffer(GL_ARRAY_BUFFER, vbonormals);
  glVertexAttribPointer(normalLoc, 3, GL_FLOAT, GL_FALSE, 0, (GLubyte *)NULL);
#endif

  // Uniform variables for the matrix
  GLint modelMatLoc = glGetUniformLocation(shaderProgram, "PVMmatrix");
  glUniformMatrix4fv(modelMatLoc, 1, GL_FALSE, glm::value_ptr(PVMmatrix)); // binding matrix to shader
  CHECK_GL_ERROR();
  // Binding vertex array object
  glBindVertexArray(vaoHandle);
  if (texobject > 0) {
    // Uniform variables for the texture
    GLint samplerID = glGetUniformLocation(shaderProgram, "texsampler");
    CHECK_GL_ERROR();
    glUniform1i(samplerID, 0); // was glUniform1i(  glUniform1f(
    CHECK_GL_ERROR();
    glActiveTexture(GL_TEXTURE0 + idTextureUnit);
    CHECK_GL_ERROR();    
    glBindTexture(GL_TEXTURE_2D, texobject);
  }
  else {
    glBindTexture(GL_TEXTURE_2D, 0);
  }
  glDrawArrays(renderMode, 0, cntVertices);
  CHECK_GL_ERROR();
  glBindVertexArray(0); // detach
  CHECK_GL_ERROR();
  glUseProgram(0);
  CHECK_GL_ERROR();
  return;
}

void
CTexturedMesh::rotate3DPoint(float angle, float *v, int axis)
{
  float sv[3];
  // ... rotation of whole enviroment sphere by angle phiShift
  int a,b;
  if(axis == 0) {
    a = 0;
    b = 1;
  }
  if(axis == 1) {
    a = 0;
    b = 2;
  }
  if(axis == 2) {
    a = 1;
    b = 2;
  }

  sv[a] = cos(angle) * v[a] - sin(angle) * v[b];
  sv[b] = sin(angle) * v[a] + cos(angle) * v[b];
  v[a] = sv[a];
  v[b] = sv[b];
  return;
}//-- rotate3DPoint -------------------------------------------------------

// rotate object geometry by an angle around axis
void
CTexturedMesh::rotateOBJModelGeometry(int axis, float rotAngle)
{
  for(int f=0;f < cntVertices; f+=3) { // for all triangles
#if 0
    // For normals, tangents, and bitangents
    int ff = f;
    for(int k=0;k<3;k++) { // for 3 vertices
      // Model rotation
      // make three points before rotation
      for(int j=0;j<3;j++) { // for xyz of a single vertex
	// For normal and tangent add the vertex and get points in 3D space
	normals[ff+j] += vertices[ff+j];
	tangents[ff+j] += vertices[ff+j];
      }
      rotate3DPoint(rotAngle, vertices[ff], axis);
      rotate3DPoint(rotAngle, normals[ff], axis);
      rotate3DPoint(rotAngle, tangents[ff], axis);
      for(int j=0;j<3;j++) { // for xyz
	N[OBJmodel][f][k][j] -= vertices[ff];
	T[OBJmodel][f][k][j] -= vertices[ff];
      }
      vec_norm(normals[ff]);
      vec_norm(tangents[ff]);
      vec_crossprod(normals[ff], tangents[ff], bitangents[ff]);
      ff += 3;
    } // for 3 vertices of a triangle
#else
    rotate3DPoint(rotAngle, &(vertices[f]), axis);    
#endif  
  } // for f
  
  glDeleteBuffers(1, &vbovertices);
  glGenBuffers(1, &vbovertices);
  glBindBuffer(GL_ARRAY_BUFFER, vbovertices);
  //glBindVertexArray(vbovertices);
  CHECK_GL_ERROR();
  int sizeVertices = 3 * sizeof(float) * cntVertices; // each vertex 3 coordinates
  glBufferData(GL_ARRAY_BUFFER, sizeVertices, vertices, GL_STATIC_DRAW);
  CHECK_GL_ERROR();
  //cout << "sizeof(vertices) = " << sizeof(vertices) << endl;
  glBindBuffer(GL_ARRAY_BUFFER, 0);
  //glBindVertexArray(0);
  return;
}//--- rotateOBJModelGeometry -----------------------------------------------

// translate object geometry by a shift in some axis
void
CTexturedMesh::translateOBJModelGeometry(int axis, float value)
{
  for(int f=0;f < cntVertices; f+=3)
  {// for faces
    vertices[f+axis] += value;
  }
  //glBindBuffer(GL_ARRAY_BUFFER, vbovertices);
  glDeleteBuffers(1, &vbovertices);
  glGenBuffers(1, &vbovertices);
  glBindVertexArray(vbovertices);
  CHECK_GL_ERROR();
  int sizeVertices = 3 * sizeof(float) * cntVertices; // each vertex 3 coordinates
  glBufferData(GL_ARRAY_BUFFER, sizeVertices, vertices, GL_STATIC_DRAW);
  CHECK_GL_ERROR();
  //cout << "sizeof(vertices) = " << sizeof(vertices) << endl;
  //glBindBuffer(GL_ARRAY_BUFFER, 0);
  glBindVertexArray(0);
  return;  
}

// -----------------------------------------------------------------------------------
// Testing code
// -----------------------------------------------------------------------------------
#if 0

// ----------------------------------------------------
// Output framebuffer data - for reading image
int fbo_width=512, fbo_height=512;
GLuint fbo=0; // frame buffer object
GLuint fbodepth=0; // depth buffer
GLuint fbotexture=0; // color buffer

// Matrices - view matrix, projection matrix, model matrix
glm::mat4 viewMatrix = glm::mat4(1.0f);
glm::mat4 projectionMatrix = glm::mat4(1.0f);
glm::mat4 modelMatrix = glm::mat4(1.0f);

// Particular meshes for testing
CTexturedMesh mesh1, mesh2;

void
initFramebuffer()
{
  glEnable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);
  // opengl setup for render buffer
  glClearColor(0.0f, 0.0f, 0.0f, 1.0f);

  // Generate color floating point texture object to be attached
  glGenTextures(1, &fbotexture); // color buffer
  cout << "fbotexture = " << fbotexture << endl;
  glBindTexture(GL_TEXTURE_2D, fbotexture);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  //glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
  //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
  //glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);
  CHECK_GL_ERROR();
  //glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE); // automatic mipmap
  // setting up texture as floating one
  glTexImage2D(GL_TEXTURE_2D,
	       0,
	       GL_RGBA32F,  // GL_RGBA32F, GL_RGBA
	       fbo_width,
	       fbo_height,
	       0,
	       GL_RGBA,
	       GL_FLOAT,
	       0);
  glBindTexture(GL_TEXTURE_2D, 0);
  CHECK_GL_ERROR();

  // ---------------------------------------------------------------------------
  // Create frame buffer object
  glGenFramebuffers(1, &fbo); // frame buffer
  CHECK_GL_ERROR();
  glBindFramebuffer(GL_FRAMEBUFFER, fbo);
  CHECK_GL_ERROR();

  glGenRenderbuffers(1, &fbodepth); // depth buffer
  CHECK_GL_ERROR();
  glBindRenderbuffer(GL_RENDERBUFFER, fbodepth);
  CHECK_GL_ERROR();
  glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT32, fbo_width, fbo_height);
  CHECK_GL_ERROR();
  glBindRenderbuffer(GL_RENDERBUFFER, 0);
  CHECK_GL_ERROR();
  
  // attach a texture to FBO color attachement point
  glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, fbotexture, 0);
  CHECK_GL_ERROR();

  // attach a renderbuffer to depth attachment point
  glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, fbodepth);
  CHECK_GL_ERROR();
  
  checkFramebufferStatus();  
  CHECK_GL_ERROR();
  
  glBindRenderbuffer(GL_RENDERBUFFER, 0);
  glClearColor(0.1,0.1,0.1,1.0);
  CHECK_GL_ERROR();
  return;
}

void
initViewMatrices()
{
  // viewing matrix
  // glm::lookAt(position, target=position+direction, up); 
  viewMatrix = glm::lookAt( glm::vec3(0.0f, 0.0f, -5.0f),
			    glm::vec3(0.0f, 0.0f, 0.0f),
			    glm::vec3(0.0f, 1.0f, 0.0f) );
#if 1
  // glm::mat4 perspective(float fovy, float aspect, float zNear,float zFar);
  projectionMatrix = glm::perspective(30.0f, 1.0f, 0.1f, 10.0f);
#else
  // orthographic projection
  const float SCENE_WIDTH = 10.0;
  const float SCENE_HEIGHT = 10.0;
  const float SCENE_DEPTH = 10.0;
  projectionMatrix = glm::ortho(
    -SCENE_WIDTH, SCENE_WIDTH,
    -SCENE_HEIGHT, SCENE_HEIGHT,
    -SCENE_DEPTH, SCENE_DEPTH
  );
#endif  

  CHECK_GL_ERROR();
  return;
}

#define FILEEXT ".exr"
void 
grabBufferHDR(char *filename, int screenShotNumber,
	      int useOrigData, int bufferFlag)
{
  // glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  float *buffer = new float[fbo_width * fbo_height * 4];
#ifdef _RENDER_TOFB
  if (bufferFlag) {
    // read from render buffer (not to the screen)
    glBindRenderbuffer(GL_RENDERBUFFER, fbo);
    glClampColor(GL_CLAMP_READ_COLOR, GL_FALSE);
    glReadBuffer(GL_COLOR_ATTACHMENT0);
    glClampColor(GL_CLAMP_READ_COLOR, GL_FALSE);
    glReadPixels(0, 0, fbo_width, fbo_height, GL_RGBA, GL_FLOAT, buffer);
    CHECK_GL_ERROR();
    glBindRenderbuffer(GL_RENDERBUFFER, 0);
    CHECK_GL_ERROR();
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    CHECK_GL_ERROR();
  }
  else {
    // from screen buffer
    glBindRenderbuffer(GL_RENDERBUFFER, 0);
    CHECK_GL_ERROR();
    glReadBuffer(GL_COLOR_ATTACHMENT0);
    CHECK_GL_ERROR();
    glReadPixels(0, 0, fbo_width, fbo_height, GL_RGBA, GL_FLOAT, buffer);
    CHECK_GL_ERROR();
  }
#else  // _RENDER_TOFB
  // read data from framebuffer of OpenGL API (visible on the screen)
  glReadBuffer(GL_BACK);
  glReadPixels(0, 0, fbo_width, fbo_height, GL_RGBA, GL_FLOAT, buffer);
#endif // _RENDER_TOFB

  // ------------------------------------------------------------------------
  //strcat(filename, FILEEXT);
  CHECK_GL_ERROR();
	   
  // writing a single file
  CHDRImage hdri;
  float fbo_width_new = fbo_width;
  hdri.allocate(fbo_width_new, fbo_height, true);    
  hdri.copyDataFromArray(fbo_width_new, fbo_height, 4, buffer, fbo_width, 0, 0, true);
  
  bool writeStatus = hdri.saveHDR(filename);
  printf("Writing framebuffer to '%s' file with signal %d "
	 "                                                                         \r",
	 filename, writeStatus ? 1:0); fflush(stdout);

  delete []buffer;
  CHECK_GL_ERROR();
  return;
}//--- grabBufferHDR -------------------------------------------------------------

void
display()
{
#if 1
  static int counter = 0;
  counter++;
  if ((counter % 2) == 0) {
    glBindFramebuffer(GL_FRAMEBUFFER, fbo);
    // glBindRenderbuffer(GL_RENDERBUFFER, 0);
  }
  else {
    glBindRenderbuffer(GL_RENDERBUFFER, 0);
  }
#endif 
  //cout << "DISPLAY" << endl << flush;
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  CHECK_GL_ERROR();
  glViewport(0, 0, fbo_width, fbo_height);
  CHECK_GL_ERROR();

  //cout << "D " << counter++ << " ";

  CHECK_GL_ERROR();
  
  // Move the model somewhere by model matrix
  glm::mat4 modelMatrix = glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 0.0f, 1.5f));
#if 1
  static float angle = 0;
  angle = ((int)(angle) + 2) % 360;
  cout << "angle " << angle << endl;
  modelMatrix = glm::rotate(modelMatrix, angle, glm::vec3(0, 1, 0)); // rotation matrix
#endif
  glm::mat4 PVMmatrix = projectionMatrix * viewMatrix * modelMatrix;

  glm::mat4 modelMatrix2 = glm::translate(glm::mat4(1.0f), glm::vec3(-1.0f, 1.0f, 1.5f));
  glm::mat4 PVMmatrix2 = projectionMatrix * viewMatrix * modelMatrix2;

#if 1
  cout << "PVMmatrix = " << endl;
  cout << PVMmatrix[0][0] << " " << PVMmatrix[1][0] << " " << PVMmatrix[2][0] << " " << PVMmatrix[3][0] << " " << endl;
  cout << PVMmatrix[0][1] << " " << PVMmatrix[1][1] << " " << PVMmatrix[2][1] << " " << PVMmatrix[3][1] << " " << endl;
  cout << PVMmatrix[0][2] << " " << PVMmatrix[1][2] << " " << PVMmatrix[2][2] << " " << PVMmatrix[3][2] << " " << endl;
  cout << PVMmatrix[0][3] << " " << PVMmatrix[1][3] << " " << PVMmatrix[2][3] << " " << PVMmatrix[3][3] << " " << endl;
  cout << "--------------------------" << endl;
  cout << "modelMatrix = " << endl;
  cout << modelMatrix[0][0] << " " << modelMatrix[1][0] << " " << modelMatrix[2][0] << " " << modelMatrix[3][0] << " " << endl;
  cout << modelMatrix[0][1] << " " << modelMatrix[1][1] << " " << modelMatrix[2][1] << " " << modelMatrix[3][1] << " " << endl;
  cout << modelMatrix[0][2] << " " << modelMatrix[1][2] << " " << modelMatrix[2][2] << " " << modelMatrix[3][2] << " " << endl;
  cout << modelMatrix[0][3] << " " << modelMatrix[1][3] << " " << modelMatrix[2][3] << " " << modelMatrix[3][3] << " " << endl;
  cout << "--------------------------" << endl;
  cout << "projectionMatrix = " << endl;
  cout << projectionMatrix[0][0] << " " << projectionMatrix[1][0] << " " << projectionMatrix[2][0] << " " << projectionMatrix[3][0] << " " << endl;
  cout << projectionMatrix[0][1] << " " << projectionMatrix[1][1] << " " << projectionMatrix[2][1] << " " << projectionMatrix[3][1] << " " << endl;
  cout << projectionMatrix[0][2] << " " << projectionMatrix[1][2] << " " << projectionMatrix[2][2] << " " << projectionMatrix[3][2] << " " << endl;
  cout << projectionMatrix[0][3] << " " << projectionMatrix[1][3] << " " << projectionMatrix[2][3] << " " << projectionMatrix[3][3] << " " << endl;
  cout << "--------------------------" << endl;
  cout << "viewMatrix = " << endl;
  cout << viewMatrix[0][0] << " " << viewMatrix[1][0] << " " << viewMatrix[2][0] << " " << viewMatrix[3][0] << " " << endl;
  cout << viewMatrix[0][1] << " " << viewMatrix[1][1] << " " << viewMatrix[2][1] << " " << viewMatrix[3][1] << " " << endl;
  cout << viewMatrix[0][2] << " " << viewMatrix[1][2] << " " << viewMatrix[2][2] << " " << viewMatrix[3][2] << " " << endl;
  cout << viewMatrix[0][3] << " " << viewMatrix[1][3] << " " << viewMatrix[2][3] << " " << viewMatrix[3][3] << " " << endl;
  cout << "===========================================" << endl;
#endif

  // Draw the mesh
  mesh1.draw(PVMmatrix);
  mesh2.draw(PVMmatrix2);
  
#if 1
  if ((counter % 2) == 0) {
    // Save it to the image
    glBindFramebuffer(GL_FRAMEBUFFER, fbo);
    CHECK_GL_ERROR();
    char filename[100];
    sprintf(filename, "test%05d.hdr", counter);
    grabBufferHDR(filename, 0, 0, true);
    //glBindFramebuffer(GL_FRAMEBUFFER, 0);
    CHECK_GL_ERROR();
    //glBindRenderbuffer(GL_RENDERBUFFER, 0);
    CHECK_GL_ERROR();
    //cout << "0";
  }
  else {
    glBindRenderbuffer(GL_RENDERBUFFER, 0);
    //cout << "1";
  }
#else
  glBindRenderbuffer(GL_RENDERBUFFER, 0);
#endif 

  glutSwapBuffers();
}

#if 1
void
idle()
{
  //static int counter2 = 0;
  //cout << "I " << counter2++;
  display();
}


void
reshape(int w, int h)
{

}

const int cntVertices2 = 6;
const int modeRenderGL2 = GL_TRIANGLES;
// Two separate triangles in XY plane - use GL_TRIANGLES
static float vertices2[cntVertices2*3] = {-1.0, -1.0, 0.0,
					-1.0, 1.0, 0.0,
					1.0, 1.0, 0.0,
					1.0, 1.0, 0.0,
					1.0, -1.0, 0.0,
					-1.0, -1.0, 0.0};
static float texcoords2[cntVertices2*2] = {0.0, 0.0,
					   0.0, 1.0,
					   1.0, 1.0,
					   1.0, 1.0,
					   1.0, 0.0,
					   0.0, 0.0};
#endif
#if 1
const int cntVertices = 4;
const int modeRenderGL = GL_TRIANGLE_STRIP;
// Two triangles in XY plane by triangle strip - use GL_TRIANGLE_STRIP
static float vertices[cntVertices*3] = {-1.0, -1.0, 0.0,
					-1.0, 1.0, 0.0,
					1.0, -1.0, 0.0,
					1.0, 1.0, 0.0};
static float texcoords[cntVertices*2] = {0.0, 0.0,
					0.0, 1.0,
					1.0, 0.0,
					1.0, 1.0};
#endif
//static float normals[12] = {0.0, 1.0, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0, 0.0};

int main(int argc, char *argv[])
{
    glutInit(&argc, argv);
    glutInitContextVersion(pgr::OGL_VER_MAJOR, pgr::OGL_VER_MINOR);
    glutInitContextFlags(GLUT_DEBUG | GLUT_FORWARD_COMPATIBLE);
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
    glutInitWindowPosition(0, 0);
    glutInitWindowSize(fbo_width, fbo_height);    
    glutCreateWindow("FBO test");
    
    glutDisplayFunc(display);
    glutIdleFunc(idle);
    glutReshapeFunc(reshape);

    if(!pgr::initialize(pgr::OGL_VER_MAJOR, pgr::OGL_VER_MINOR))
      pgr::dieWithError("pgr init failed, required OpenGL not supported?");
    // create objects to be rendered
    mesh1.init(cntVertices, modeRenderGL, vertices, texcoords);
    mesh1.loadTexture("ENV/stpeters.hdr");

    mesh2.init(cntVertices2, modeRenderGL2, vertices2, texcoords2);
    mesh2.loadTexture("ENV/uffizi_lalo.hdr");

    glClampColor(GL_CLAMP_READ_COLOR, GL_FALSE);
    initFramebuffer();
    initViewMatrices();

    glutMainLoop();

    return 0;
}

#endif
