// ================================================================
// $Id: basrnd.h,v 1.5 2005/11/30 15:32:41 havran Exp $
//
// basrnd.h
//     Basic operations for random generator.
//
// REPLACEMENT_STRING
//
// Initial coding by Vlastimil Havran, 2004.

#ifndef __BASRND_H__
#define __BASRND_H__

#include "configh.h"
#include "basmath.h"

// Standard headers
#include <cstdlib>

// forward declarations

/// Returns a random floating-point value between the two values. 
/**
  Range is inclusive; the function should occasionally return a, 
  but never b. The range is [a; b).
*/
inline float
RandomValue(float a, float b)
{
  float range = Abs(a - b);

// use Mersenne Twister random number generator
#define __USE_MERSENNE_TWISTER

#ifdef __USE_MERSENNE_TWISTER
  // preferred way
  extern double MersenneTwisterRand();
  return MersenneTwisterRand() * range + ((a < b) ? a : b);
#else
#ifdef __UNIX__
  // traditional ANSI generator - the worst case
  return drand48() * range + ((a < b) ? a : b);
#else // __UNIX__
  const double MULT = 1.0 / (RAND_MAX+1);
  return ((float)rand() * MULT) * range + ((a < b) ? a : b);
#endif // __UNIX__

#endif // __USE_MERSENNE_TWISTER
}

/// Returns random value in the range [0; 1)
inline float
RandomValue()
{
#ifdef __USE_MERSENNE_TWISTER
  // preferred way
  extern double MersenneTwisterRand();
  return MersenneTwisterRand();
#else
#ifdef __UNIX__
  // traditional ANSI generator - the worst case  
  return drand48();
#else // __UNIX__
  const double MULT = 1.0 / (RAND_MAX + 1);
  return rand() * MULT;
#endif // __UNIX__
#endif // __USE_MERSENNE_TWISTER
}


/// Sets the random seed for all random generators.
inline void
SetRandomSeed(int seed = 0)
{
#ifdef __USE_MERSENNE_TWISTER
  // preferred way
  extern void MersenneTwisterSetSeed(int);
  MersenneTwisterSetSeed(seed);
#endif // #__USE_MERSENNE_TWISTER

  srand(seed);
#ifdef __UNIX__
  srand48(seed);
#endif // __UNIX__
}

#endif // __BASRND_H__

