// ================================================================
// $Id: $
//
// render1.cpp
//
// This implements deterministic sampling from environment map, all pixels the same set of lights
//
// Initial coding by Vlastimil Havran, 2013.

#include <cassert>
#include <cmath>

#include "configg.h"
#include "AuxFuncs.h"
//#include "basrnd.h"
#include "ibtfbase.h"
#include "lightenv.h"
#include "brdfmodel.h"
#include "vrmlsup.h"
#include "tangles.h"
#include "render.h"

// ================================================================================
// Deterministic coherent importance sampling of environment map of
// a single pixel, standard mode either point light or single
// coherent samples at environment map - this is destuingihsed by switch
void         // interpolated vertices, tangents, normals and binormals
computeMode1(float Vint[3], float Tint[3], float Nint[3], float Bint[3], 
	     float t_v, float p_v, int irow, int jcol, int BTFindex, float normF)
{
  float t_i, p_i;
  float RGB[3], result[3];
  // cout << "A" << std::flush;
  result[0] = result[1] = result[2] = 0.f;
  resetCountersST();
  //cout << " normF = " << normF << endl;
  
  // counting samples
  int failEM = 0;
  int cntAllSamplesEM = 0;
  int cntValidSamplesEM = 0;
  // -----------------------------------------------------------------------------
  // for all samples
  const bool allowNewSamples = false;
  int samplesMax = cntSamplesPerPixel;
  for(int l=0; l < samplesMax; l++) {
    // copy the positions of the directional light for environment map here
    // this direction of light was precomputed
    assert(cntAllSamplesEM < maxCntSamplesPerPixel);
    cntAllSamplesEM++;
    // possibly randomize the access to light
    environMap[actualEnvMap].generateNewRandomLightPrecomputed();
    TA->setIlluDir(environMap[actualEnvMap].getDirectionPrecomputedLight(l)); 
    TA->computeIlluAngles(Vint, Tint, Bint, Nint, &t_i, &p_i);
    // theta and phi are in radians
    if (t_i >= M_PI/2.f) {
      // nothing to do with the illumination from this light
      failEM++;
      if ((allowNewSamples) && (samplesMax < maxCntSamplesPerPixel)) {
	// take additional samples
	samplesMax++;
      }
      continue;
    }
    cntValidSamplesEM++;
    // the value given by BRDF model
    float brdfRGB[3];
    // the cos(theta) factor is already included in the call below
    bool tt = brdf->getBRDFValue(irow, jcol, t_v, p_v, t_i, p_i, brdfRGB);
    // this value has been precomputed
    float emRGB[3];
    for (int isp=0; isp < 3; isp++)
      emRGB[isp] = environMap[actualEnvMap].getColorPrecomputedLight(l, isp);
    float probEM = luminance(emRGB);

    //float probEM = environMap[actualEnvMap].getRGBandProb_ThetaPhi(t_i, p_i, envRGB);
    //float probEM2 = environMap[actualEnvMap].getProb_ThetaPhi(t_i, p_i);
    //float probEM = luminance(envRGB)/(4.0f*M_PI);

    for(int isp=0;isp<=2;isp++) {// for all 3 color channels
      result[isp] +=
	// the colour of light from EM was precomputed
	brdfRGB[isp] * environMap[actualEnvMap].getColorPrecomputedLight(l, isp)/probEM;
    } // for all channels
  } // for all lights
  failEMcnt += failEM;
  evalEMcnt += cntAllSamplesEM;
  // The number of samples taken is cntAllSamplesEM
  // The number of failedSampled is failEM
  // The number of success Samples is cntAllSamplesEM-failEM

  //  float lumEM = luminance(result)/(float)cntAllSamplesEM;
  //  cout << "EM " << lumEM << endl;
  
  float multIntensity = normF / (float)cntAllSamplesEM;
  for(int isp=0;isp<=2;isp++) {
    BTFimageF[BTFindex+isp] = result[isp] * multIntensity;
  }
  setCountersST(0, cntAllSamplesEM);
  return;
}

