// ===================================================================
// $Id: $
//
// tangles.h
//
// Class: TAngles
//
//  The class to implement rendering via OpenGL with shaders and HDR texture
//
// Initial coding by Jiri Filip, 2006.

#ifndef __TANGLES_H__
#define __TANGLES_H__

class TAngles
{
private:
  int NoOfDirections;         //! no. of view/illu. measurement per hemisphere in UBO data
  float **angles;             //! set of measurements directions of UBO BTF datasets

  float t_i, p_i;              //! elevation and azimuthal angles of illumination direction
  float t_v, p_v;              //! elevation and azimuthal angles of viewing direction 
  float *i_weight, *v_weight; //! 3 barycentric texture interpolation weights
  int *i_index, *v_index;     //! 3 closest indices to illumination direction and viewing direction

public:
  TAngles();
  ~TAngles();

  float *texWeights;          //! 9 barycentric texture interpolation weights
  int *illuPos, *viewPos;     //! 9 closest indices to illumination and viewving directions
  float *c_pos;               //! camera position
  float *l_pos;               //! illumination position

  int print;                  //! flag for printing debugging info

  // generate set of measurements directions for UBO BTF datasets
  void getViewPos(float vpos[]);
  void getIlluPos(float ipos[]);

  void setViewPos(float vpos[]);
  void setIlluPos(float ipos[]);
  void setIlluDir(float ipos[]);
  
  void generateDirections();
  void computeTB(float **v, float **UV, float T[], float B[]);
  int computeViewAnglesSphere(float *v, float *N);
  void computeIlluAnglesSphere(float *v, float *N);
  int computeAnglesSphere(float *v, float *N);          

  int computeViewAngles(float V[], float T[], float B[], float N[],
			float *theta_v, float *phi_v);
  void computeIlluAngles(float V[], float T[], float B[], float N[],
			 float *theta_i, float *phi_i);

  void getAngles(float *theta_i, float *phi_i, float *theta_v, float *phi_v);
};

#endif // __TANGLES_H__
