// ================================================================
// $Id: $
//
// vrmlsup.cpp
//
// Functions to make easy debugging via VRML'97 file format.
//
// REPLACEMENT_STRING
//
// Initial coding by Vlastimil Havran, 2005.

#include <cstdlib>
#include <iostream>
#include <fstream>
#include <cmath>
#include "vrmlsup.h"

// The stream to put VRML file for 3D visualization
static std::ofstream VRMLStream;

void
VRMLStreamPutPoint(
  const double x, double y, double z,
  double r, double g, double b,
  double radius)
{
#if 1
  // representation via spheres 
  VRMLStream << "Transform {\n";
  VRMLStream << "  translation " << x << " " << y << " " << z << "\n";    
  VRMLStream << "  children[\n";
  VRMLStream << "    Shape {\n";
  VRMLStream << "      appearance Appearance {\n";
  VRMLStream << "        material Material {\n";
  VRMLStream << "          diffuseColor "
	    << r/256.0 << " " << g/256.0 << " " << b/256.0 << "\n";
  VRMLStream << "        }\n";
  VRMLStream << "      }\n"; // appearance
  VRMLStream << "      geometry Sphere { radius " << radius << "}\n";
  VRMLStream << "    }\n"; // shape
  VRMLStream << "  ]\n"; // children
  VRMLStream << " }\n"; // transform
#else
  VRMLStream << "Shape {\n";
  VRMLStream << "  geometry PointSet { coord Coordinate { point [\n";
  VRMLStream << x << " " << y << " " << z << ", \n";
  VRMLStream << "] } color Color { color [\n";
  VRMLStream << r/256.0 << " " << g/256.0 << " " << b/256.0 << " ,\n";
  VRMLStream << "] } } }\n";
#endif  
  // to be sure that it is written to a file
  VRMLStream << std::endl << std::flush;
  
  return;
}

void
VRMLStreamOpen(char *filename)
{
  // and init the stream - rewrite
  VRMLStream.open(filename, std::ios::trunc | std::ios::binary);

  if (!VRMLStream) {
    std::cerr << "Opening of stream failed" << std::endl;
    exit(5);
  }

  VRMLStream << "#VRML V2.0 utf8\n" << std::endl ;

  VRMLStream << "Viewpoint {\n";
  VRMLStream << "  position 0 0 3.0\n";
  VRMLStream << "  orientation 0 0 1 0\n";
  VRMLStream << "  fieldOfView 0.785398\n";
  VRMLStream << "}" << std::endl;
}

void
VRMLStreamClose()
{
  // to be sure that it is written to a file
  VRMLStream << std::endl << std::flush;
  VRMLStream.close();
}

void
VRMLStreamSaveAxes(float cx, float cy, float cz, float length, double radius)
{
  int cnt = length/(2.0f*radius);
  if (cnt > 1000) cnt = 1000;
  VRMLStream << std::endl;
  VRMLStream << "# Saving axes visualization by << " << cnt
	     << " spheres of red(x-axis), green(y-axis), blue(z-axis)." << std::endl << std::flush;

  for (int i = 0; i < cnt; i++) // x-axis is red
    VRMLStreamPutPoint(cx + (i*2+1)*radius, cy, cz, 255.0, 0.0, 0.0, radius);
  for (int i = 0; i < cnt; i++) // y-axis is green
    VRMLStreamPutPoint(cx, cy + (i*2+1)*radius, cz, 0.0, 255.0, 0.0, radius);
  for (int i = 0; i < cnt; i++) // z-axis is blue
    VRMLStreamPutPoint(cx, cy, cz + (i*2+1)*radius, 0.0, 0.0, 255.0, radius);
  return;
}

void
VRMLStreamSaveLine(float cx, float cy, float cz,
		   float dx, float dy, float dz,
		   float red, float green, float blue,
		   float length, float radius)
{
  int cnt = length/(2.0f*radius);
  if (cnt > 1000) cnt = 1000;
  VRMLStream << std::endl;
  VRMLStream << "# Saving axes visualizatio by << " << cnt
	     << " points." << std::endl << std::flush;
  float mag = sqrt(dx*dx+dy*dy+dz*dz);
  dx/=mag;
  dy/=mag;
  dz/=mag;
  float radius2 = radius;
  for (int i = 0; i < cnt; i++) {// given a color
    if (i > cnt-3) radius2 *=1.5;
    VRMLStreamPutPoint(cx + (i*2+1)*radius*dx,
		       cy + (i*2+1)*radius*dy,
		       cz + (i*2+1)*radius*dz,
		       red, green, blue, radius2);
  } // for i
  return;
}
