// ================================================================
// $Id: basstr.h,v 1.4 2006/09/29 09:12:39 Exp $
//
// basstr.h
//     Basic operations with strings
//
// Initial coding by Vlastimil Havran, 2000.

#ifndef __BASSTR_H__
#define __BASSTR_H__

// GOLEM headers
#include "configh.h"

// Standard headers
#include <string>

__BEGIN_SBRDFLIB_HEADER

// forward declarations

#define DIRCAT 0x2f

// ----------------------------------------------------------------------
// String functions for filenames

/// Puts the specified number of spaces out to given stream.
/** 
    Used for indentation during CObject3D::Describe().
*/
void
indent(ostream &app, int ind);

/// Prefix filename by a given directory name. 
/**
   E.g. "aa.out","ref" makes "ref/aa.out", but "stats/aa.out",
   "ref" makes "stats/ref/aa.out".
 */
void
PrefixNameByDir(const string &filename, const string &dirname,
		string &outputName);

/// Prefix filename extension by a given extPrefix.
/**
   For example the filename "data/tetra.rgb" and extPrefix "0102" makes
   the output filename "data/tetra0102.rgb"
   If no extension is found, then extPrefix is simply appended to the
   file name, e.g. for "balls" makes "balls0102"
*/
void
PrefixExtensionByName(const string &filename, const string &extPrefix,
		      string &outputName);

/// A version of PrefixExtensionByName which returns the result by value.
inline string 
PrefixExtensionByName(const string &filename, const string &extPrefix)
{
  string outputName;
  PrefixExtensionByName(filename,extPrefix,outputName);
  return outputName;
}

/// Adding/replacing the extension of the filename.
/**
   @param filename  (in) filename to be changed, e.g. "tetra.rgb"
   @param newExtension  (in) newExtension added/replaced  (e.g. "vis", "dbg",
                             "rgb" etc.)
   @param outputName  (out) filename with replaced extension 

   (e.g. "tetra.rgb" makes "tetra.dbg" and "aaaaa/bbb" makes "aaaaa/bbb.dbg").
*/
void
ChangeFilenameExtension(const string &filename, const string &newExt,
			string &outputName);

/// A version of ChangeFilenameExtension which returns the result by value.
inline string
ChangeFilenameExtension(const string &filename, const string &newExt)
{
  string outputName;
  ChangeFilenameExtension(filename,newExt,outputName);
  return outputName;
}

void
ExtractFilenameExtension(const string &filename, string &xtension);

/// A version of ChangeFilenameExtension which returns the result by value.
inline string
ExtractFilenameExtension(const string &filename)
{
  string outputName;
  ExtractFilenameExtension(filename,outputName);
  return outputName;
}

// Removes path from filename,
// for example the filename "data/tetra.rgb" gets "tetra.rgb"
void
RemovePathFromFilename(const string &filename, string &outputName);

/**
  Skip all the rows in a text file starting from the current character,
  when the current character is a 'mark'. All the following lines starting
  with the 'mark' character are also skipped.
*/
void
EatComments(istream &inpf, const char mark);

/// Skip all the whitespaces in the input stream. 
/**
   Whitespaces are the characters ' ','\t','\n','\f','\r','\b' .
*/
void
EatWhitespace(istream &inpf);


// Find the first line from now on that starts with a given keyword.
/**
   Read the keyword and do not read other characters
   Returns 'false' on success (line found), 'true' if not found.
*/
bool
FindLineWithKeyword(istream &inpf, const string &keyword);

/// Converts all the possible characters to lowercase.
void
ToLowerCase(string &str);

__END_SBRDFLIB_HEADER

#endif // __SBRDFLIB_H__

